package at.jku.ssw.psw2.demos.db;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

public class PersonDBManager {
	
	/**
	 * static constructor loading the driver class and creating the singleton instance. 
	 */
	static {
		try {
			Class.forName("org.apache.derby.jdbc.EmbeddedDriver");
			getInstance(); 
		} catch (SQLException | ClassNotFoundException e) {
			
		}
	}

	/** Constant string for the database url */ 
	private static final String DBURL = "jdbc:derby:./personsdb"; 

	/** the variable holding the singleton instance */
	private static PersonDBManager INSTANCE; 
	
	/**
	 * Accesses the singleton instance
	 * @return the singleton instance
	 * @throws SQLException
	 */
	public synchronized static PersonDBManager getInstance() throws SQLException {
		if (INSTANCE == null) {
			INSTANCE = new PersonDBManager(); 
		}
		return INSTANCE;
	}
	
	private Connection conn; 
	private PreparedStatement insertPersonStmt; 
	private PreparedStatement updatePersonStmt; 
	private PreparedStatement selectPersonByIdStmt; 
	private PreparedStatement selectPersonByNameStmt; 
	private PreparedStatement selectPersonsStmt; 
	private PreparedStatement deleteAllStmt; 
	
	/** 
	 * Private constructor connecting to database and creating the database statements. 
	 * @throws SQLException
	 */
	private PersonDBManager() throws SQLException {	
		try {
			conn = DriverManager.getConnection(DBURL);
		} catch (Exception e) {
			conn = DriverManager.getConnection(DBURL + ";create=true");
			createTables(conn); 
		}
		insertPersonStmt = conn
				.prepareStatement("insert into Persons values(DEFAULT, ?, ?)");
		updatePersonStmt = conn
				.prepareStatement("update Persons set name=?, sex=? where id=?");
		selectPersonsStmt = conn
				.prepareStatement("select * from Persons");
		selectPersonByNameStmt = conn
				.prepareStatement("select * from Persons where name=?");
		selectPersonByIdStmt = conn
				.prepareStatement("select * from Persons where id=?");
		deleteAllStmt = conn
				.prepareStatement("delete from Persons");
	}
	
	/**
	 * Inserts a new person record into the Persons table. 
	 * @param name the name of the person 
	 * @param sex the sex of the person
	 * @return the person object representing the record created 
	 * @throws SQLException
	 */
	public Person insertPerson(String name, Sex sex) throws SQLException {
		synchronized (insertPersonStmt) {
			insertPersonStmt.setString(1, name);
			insertPersonStmt.setInt(2, (sex == Sex.Male ? 0 : 1));
			insertPersonStmt.executeUpdate();
			return getPersonByName(name); 
		}
	}
	
	/**
	 * Updates the data of a person with the given id. 
	 * @param id the id of the person (used as key in the Persons table)
	 * @param name the new name of the person 
	 * @param sex the possibly change sex of the person 
	 * @return the person object with the changed values 
	 * @throws SQLException
	 */
	public Person updatePerons(int id, String name, Sex sex) throws SQLException {
		synchronized (updatePersonStmt) {
			updatePersonStmt.setString(1, name);
			updatePersonStmt.setInt(2, (sex == Sex.Male ? 0 : 1));	
			updatePersonStmt.setInt(3, id);
			updatePersonStmt.executeUpdate();
			return getPersonById(id); 
		}
	}
	
	/**
	 * Returns the array of the person object from the records in the Person database. 
	 * Uses a select * statement to retrieve all data from the Persons table.  
	 * @return the array of person objects 
	 * @throws SQLException
	 */
	public Person[] getPersons() throws SQLException {
		List<Person> persons = new ArrayList<>(); 
		ResultSet r = selectPersonsStmt.executeQuery(); 
		while (r.next()) {
			Person p = new Person(r.getInt(1), r.getString(2), 
					r.getInt(3) == 0 ? Sex.Male : Sex.Female);
			persons.add(p);
		}
		return persons.toArray(new Person[persons.size()]);
	}
	
	/**
	 * Gets the person object from the given name of a person. 
	 * Reads the person with given name from the Persons table.  
	 * @param name the name of the person to read 
	 * @return the person object with that name
	 * @throws SQLException
	 */
	public Person getPersonByName(String name) throws SQLException {
		ResultSet r;
		synchronized (selectPersonByNameStmt) {
			selectPersonByNameStmt.setString(1, name);
			r = selectPersonByNameStmt.executeQuery(); 
		}
		if (r.next()) {
			return new Person(r.getInt(1), r.getString(2), r.getInt(3) == 0 ? Sex.Male : Sex.Female);
		} else { 
			return null; 
		}
	}
	
	/**
	 * Gets the person object from the given id of a person. 
	 * Reads the person with given id from the Persons table.  
	 * @param id the unique id of the person 
	 * @return the person object with that id
	 * @throws SQLException
	 */
	public Person getPersonById(int id) throws SQLException {
		ResultSet r;
		synchronized (selectPersonByIdStmt) {
			selectPersonByIdStmt.setInt(1, id);
			r = selectPersonByIdStmt.executeQuery(); 
		}
		if (r.next()) {
			return new Person(r.getInt(1), r.getString(2), r.getInt(3) == 0 ? Sex.Male : Sex.Female);
		} else { 
			return null; 
		}
	}
	
	/**
	 * Deletes all records from the Persons table. 
	 * @throws SQLException
	 */
	public void deleteAll() throws SQLException {
		deleteAllStmt.execute(); 

	}
	
	/**
	 * Closes the database connection. 
	 * @throws SQLException
	 */
	public void close() throws SQLException {
		conn.close();
	}

	/**
	 * Creates the Persons table. 
	 * @param conn
	 */
	private static void createTables(Connection conn) {
		String createPersonsTableStr = 
				"create table Persons (id integer primary key generated always as identity," + 
                "name varchar(30), sex int default 0 )";  
		try {
			PreparedStatement createPersonsTableStmt = conn.prepareStatement(createPersonsTableStr);
			createPersonsTableStmt.executeUpdate();
		} catch (SQLException e) {
			e.printStackTrace();
		} 
	}
	
}
